/*
 * Copyright 
 *   2007 axYus - www.axyus.com
 *   2007 JP.Tessier - jean-philippe.tessier@axyus.com
 * 
 * This file is part of XEMELIOS.
 * 
 * XEMELIOS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * XEMELIOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with XEMELIOS; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package fr.gouv.finances.cp.xemelios.controls.models;

import java.util.Hashtable;
import java.util.Vector;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import fr.gouv.finances.cp.utils.xml.marshal.InvalidXmlDefinition;
import fr.gouv.finances.cp.utils.xml.marshal.XmlAttributes;
import fr.gouv.finances.cp.utils.xml.marshal.XmlMarshallable;
import fr.gouv.finances.cp.utils.xml.marshal.XmlOutputter;
import javax.xml.namespace.QName;


public class ControlModelTech implements XmlMarshallable{
    public static final transient String TAG = "controlTech";
    public static final transient QName QN = new QName(TAG);
    public String id, libelle, severity;
    public Vector<ParamModelTech> params;
    private Hashtable<String,ParamModelTech> indexedParamsTech;
    
    public ControlModelTech(QName tagName) {
        super();
        params=new Vector<ParamModelTech>();
        indexedParamsTech=new Hashtable<String,ParamModelTech>();
    }
    
    public void addCharacterData(String cData) throws SAXException { }
    
    public void addChild(XmlMarshallable child, QName tagName) throws SAXException {
        if(ParamModelTech.QN.equals(tagName)) {
            ParamModelTech cm = (ParamModelTech)child; 
            if(indexedParamsTech.containsKey(cm.getId())) {
                ParamModelTech old = indexedParamsTech.get(cm.getId());
                params.remove(old);
                indexedParamsTech.remove(old.getId());
            }
            params.add(cm);
            indexedParamsTech.put(cm.getId(),cm);
        }
    }
    
    public XmlMarshallable getAttributes(XmlAttributes attributes) throws SAXException {
        id=attributes.getValue("id");
        libelle=attributes.getValue("libelle");
        severity=attributes.getValue("severity");
        return this;
    }
    public void validate() throws InvalidXmlDefinition {
        for(ParamModelTech pcm:params) pcm.validate();
    }
    public void marshall(XmlOutputter output) {
        output.startTag(TAG);
        output.addAttribute("id",id);
        output.addAttribute("libelle",libelle);
        output.addAttribute("severity",severity);
        for(ParamModelTech pcm:params) pcm.marshall(output);
        output.endTag(TAG);
    }
    
    public String getId() {return id;}
    public String getLibelle() {return libelle;}
    public String getSeverity() {return severity;}
    public Vector<ParamModelTech> getParams() {return params;}
    
    public ParamModelTech getParamTechById(String ctrlId) {
        return indexedParamsTech.get(ctrlId);
    }
    
    public ControlModelTech clone() {
        ControlModelTech cm = new ControlModelTech(QN);
        cm.id = this.id;
        cm.libelle = this.libelle;
        cm.severity = this.severity;
        for(ParamModelTech pmt:getParams()) {
        	try {
        		addChild(pmt.clone(), ParamModelTech.QN);
        	} catch(SAXException ignore) {}
        }
        return cm;
    }

    public XmlMarshallable getChildToModify(String uri, String localName, String qName, Attributes atts) {
        if(ParamModelTech.QN.equals(new QName(uri,localName))) {
            return indexedParamsTech.get(atts.getValue("id"));
        }
        return null;
    }

    public QName getQName() {
        return QN;
    }
}
